SystemOrganization addCategory: #'Ma client server-API'!
SystemOrganization addCategory: #'Ma client server-Info'!
SystemOrganization addCategory: #'Ma client server-Exceptions'!

!ByteArray methodsFor: '*ma client server' stamp: 'cmm 10/1/2002 19:42'!
maAsIpString

	^self size = 4
		ifTrue:
			[ (WriteStream on: (String new: 4))
				maPrint: (self at: 1);
				nextPut: $.;
				maPrint: (self at: 2);
				nextPut: $.;
				maPrint: (self at: 3);
				nextPut: $.;
				maPrint: (self at: 4);
				contents ]
		ifFalse: [ '(not an ip address)' ]! !

MaUserError subclass: #MaClientServerUserError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Exceptions'!

MaObject subclass: #MaClientServerPreferences
	instanceVariableNames: ''
	classVariableNames: 'Debug'
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaClientServerPreferences commentStamp: 'cmm 2/1/2005 15:27' prior: 0!
Control your Ma client server preferences with this class.!

!MaClientServerPreferences class methodsFor: 'accessing' stamp: 'cmm 1/2/2003 14:35'!
debug

	^Debug! !

!MaClientServerPreferences class methodsFor: 'accessing' stamp: 'cmm 1/2/2003 14:53'!
debug: aBoolean

	Debug _ aBoolean! !

!MaClientServerPreferences class methodsFor: 'private' stamp: 'cmm 1/2/2003 14:35'!
initialize

	super initialize.
	Debug _ false! !

MaObject subclass: #MaClientSocket
	instanceVariableNames: 'hostAddress port socket byteArray timeoutSeconds'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaClientSocket commentStamp: 'cmm 2/1/2005 15:29' prior: 0!
An easy-to-use class to send ByteArray or String requests to a socket listening on the network.  The other socket is listening using MaServerSocket.!

!MaClientSocket class methodsFor: 'creation' stamp: 'cmm 8/28/2003 22:57'!
hostAddress: aByteArray port: anInteger

	Socket initializeNetwork.
	^self new
		hostAddress: aByteArray;
		port: anInteger! !

!MaClientSocket class methodsFor: 'initializing' stamp: 'cmm 5/2/2003 15:08'!
initialize

	Smalltalk addToStartUpList: self! !

!MaClientSocket class methodsFor: 'initializing' stamp: 'cmm 5/2/2003 15:11'!
startUp: aBoolean

	super startUp: aBoolean.
	Socket initializeNetwork! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 9/22/2002 21:33'!
disconnect

	socket close! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 11/26/2004 09:45'!
ensureByteArrayCanHold: anInteger

	anInteger > 10000000 ifTrue: [ MaSecurityViolation signal: 'Possible attack.' ].
	anInteger > byteArray size
		ifTrue:
			[ | newByteArray |
			newByteArray _ ByteArray new: anInteger.
			newByteArray
				replaceFrom: 1
				to: byteArray size
				with: byteArray
				startingAt: 1.
			byteArray _ newByteArray ]! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 9/20/2002 14:28'!
headerSize

	^4! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 9/20/2002 12:24'!
hostAddress

	^hostAddress! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 9/20/2002 12:23'!
hostAddress: aByteArray

	hostAddress _ aByteArray! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 10/1/2002 19:43'!
hostAndPortString

	^'host: ', hostAddress maAsIpString, ' port: ', port printString! !

!MaClientSocket methodsFor: 'initializing' stamp: 'cmm 12/24/2002 15:49'!
initialize

	super initialize.
	byteArray _ ByteArray new: 1000.
	timeoutSeconds _ 30! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 1/27/2003 23:57'!
loadByteArrayWith: aByteArray startingAt: indexPosition count: anInteger

	self ensureByteArrayCanHold: self headerSize + anInteger.
	byteArray
		replaceFrom: self headerSize + 1
		to: self headerSize + anInteger
		with: aByteArray
		startingAt: indexPosition.
	byteArray
		maUint: 32
		at: 0
		put: self headerSize + anInteger! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 9/20/2002 12:23'!
port

	^port! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 9/20/2002 12:23'!
port: anInteger

	port _ anInteger! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 1/14/2005 10:38'!
receiveInto: aByteArray
	"First, we receive into *my* byteArray, then I paste it into aByteArray."

	| bytesExpected bytesReceived answer |
	bytesReceived _ 0.
	bytesExpected _ nil.
	[ bytesExpected isNil or: [ bytesReceived < bytesExpected ] ]
		whileTrue:
			[ socket 
				waitForDataFor: timeoutSeconds
				ifClosed: [ MaNetworkError signal: 'Connection closed while waiting for response.' ]
				ifTimedOut: [ MaNetworkError signal: 'response not received from ', self hostAndPortString, ' within the timeout period.' ].
			bytesReceived _ bytesReceived
				+ (socket
					receiveSomeDataInto: byteArray
					startingAt: bytesReceived + 1).
			bytesExpected ifNil:
				[ bytesReceived >= self headerSize 
					ifTrue:
						[ bytesExpected _ byteArray maUint: 32 at: 0.
						self ensureByteArrayCanHold: bytesExpected ] ] ].
	"We should now have the full buffer in my byteArray, paste it into the answer ByteArray."
	answer _ (bytesReceived - self headerSize) > aByteArray size
		ifTrue: [ ByteArray new: bytesReceived - self headerSize ]
		ifFalse: [ aByteArray ].
	answer
		replaceFrom: 1
		to: bytesReceived - self headerSize
		with: byteArray
		startingAt: self headerSize + 1.
	^ answer! !

!MaClientSocket methodsFor: 'private' stamp: 'cmm 7/2/2004 16:07'!
send: anInteger

	socket
		sendData: byteArray
		count: anInteger.
	socket waitForSendDoneFor: timeoutSeconds.
	socket sendDone ifFalse: [ MaNetworkError signal: 'Connection to ', self hostAndPortString, ' lost.' ]! !

!MaClientSocket methodsFor: 'send' stamp: 'cmm 1/30/2003 22:08'!
sendData: requestByteArray

	| answer |
	answer _ ByteArray new: 500.
	^self
		sendData: requestByteArray
		startingAt: 1
		count: requestByteArray size
		waitForReplyIn: answer! !

!MaClientSocket methodsFor: 'send' stamp: 'cmm 3/30/2005 14:42'!
sendData: requestByteArray startingAt: startIndex count: numberOfBytes waitForReplyIn: answerByteArray
	"Perform a synchronous TCP call, sending count bytes of requestByteArray from position startIndex to my peer socket and then wait up to my timeout period for a reply, which I'll put in answerByteArray.  If I run out of room in answerByteArray, I'll create you a new one and return it, otherwise you'll get your own answerByteArray back.  If communications are disrupted, I'll signal a MaNetworkError."

	| answer retries |
	socket _ Socket newTCP.
	retries _ 0.
	[ socket 
		connectTo: hostAddress 
		port: port ]
		on: NetworkError
		do: 
			[ : error | 
			retries = 0
				ifTrue: [ retries _ retries + 1.  error retry ]
				ifFalse: [ MaNetworkError signal: error messageText ] ].
	self
		loadByteArrayWith: requestByteArray
		startingAt: startIndex
		count: numberOfBytes.
	[ self send: numberOfBytes + self headerSize ]
		on: NetworkError
		do: [ : err | MaNetworkError signal: err messageText, ' while sending request.' ].
	[ answer _ self receiveInto: answerByteArray ]
		on: NetworkError
		do: [ : err | MaNetworkError signal: err messageText, ' while receiving response.' ].
	socket closeAndDestroy: timeoutSeconds.
	^ answer! !

!MaClientSocket methodsFor: 'accessing' stamp: 'cmm 9/20/2002 15:10'!
timeoutSeconds: anInteger

	timeoutSeconds _ anInteger! !

MaObject subclass: #MaServer
	instanceVariableNames: 'console'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaServer commentStamp: 'cmm 2/1/2005 15:35' prior: 0!
Abstract superclass for the "server engine".  I know a console which I send output to.!

MaServer subclass: #MaRequestServer
	instanceVariableNames: 'processor'
	classVariableNames: 'ActiveServers'
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaRequestServer commentStamp: 'cmm 2/20/2005 17:15' prior: 0!
Abstract class whose subclass instances handle the server portion of the client-server activity.  The processor should be an object that responds to #value:.  The argument is the actual request object and it should answer the result object.  Requests and responses can be arbitrary objects, not just byte-objects like Strings and ByteArrays, although you must set up your #protocol.!

MaRequestServer subclass: #MaLocalRequestServer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaLocalRequestServer commentStamp: 'cmm 2/1/2005 15:29' prior: 0!
Use this server as a stub when you are running your server locally within the same image.!

!MaLocalRequestServer class methodsFor: 'creation' stamp: 'cmm 9/19/2004 22:38'!
processor: oneArgBlockOrProcessor

	^ self new 
		processor: oneArgBlockOrProcessor ;
		yourself! !

!MaRequestServer methodsFor: 'initializing' stamp: 'cmm 1/5/2003 23:12'!
initialize

	super initialize.
	console _ MaRequestServerConsole new! !

!MaRequestServer methodsFor: 'testing' stamp: 'cmm 2/20/2005 17:18'!
isRunning

	self subclassResponsibility! !

!MaRequestServer methodsFor: 'private' stamp: 'cmm 9/19/2004 22:09'!
processRequest: aMaServerRequest
	"Here's your request object, do something with it and answer any object, though for remote connections only, every class throughout the entire request and responses graph must be part of your #protocol."

	| answer |
	console noteProcessingRequest: aMaServerRequest.
	answer _ [ processor value: aMaServerRequest ]
		on: MaNotification
		do:
			[ :noti | console note: noti messageText.
			noti resume ].
	console noteRequestProcessed: aMaServerRequest.
	^answer! !

!MaRequestServer methodsFor: 'accessing' stamp: 'cmm 12/8/2004 13:34'!
processor
	^ processor! !

!MaRequestServer methodsFor: 'private' stamp: 'cmm 9/19/2004 22:10'!
processor: oneArgBlock
	"oneArgBlock takes the request object, answers the response object you want to send back to the client.  For remote, these can only be objects defined in the protocol."

	processor _ oneArgBlock! !

MaRequestServer subclass: #MaTcpRequestServer
	instanceVariableNames: 'socket serializer nextId'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaTcpRequestServer commentStamp: 'cmm 2/1/2005 15:40' prior: 0!
The server to use when you want to allow requests and responses to be any object and run on a remote server.!

!MaTcpRequestServer class methodsFor: 'creation' stamp: 'cmm 10/8/2002 22:50'!
protocol: collectionOfClasses
	"collectionOfClasses are all request classes and all response classes any client can expect to encounter in the full graph of the requests and responses"

	^self new protocol: collectionOfClasses! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 10/11/2002 00:24'!
connectionInfo

	^Array
		with: self getNextId
		with: self protocolByteArray! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 10/11/2002 00:22'!
connectionRequest

	^'connect'! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 10/11/2002 00:25'!
getNextId

	^nextId _ nextId + 1! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 11/26/2004 13:15'!
initialize

	super initialize.
	socket _
		MaServerSocket new
			server: self;
			yourself.
	nextId _ 1.
	self initializeSerializer! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 2/6/2005 12:37'!
initializeSerializer
	serializer _ MaObjectSerializer new.
	serializer preferences
		beforeSerializingAny: #Exception
		do:
			[ : anException |
			#('signalContext' 'handlerContext' 'outerContext') do:
				[ : each |
				anException
					instVarNamed: each
					put: nil ] ]! !

!MaTcpRequestServer methodsFor: 'testing' stamp: 'cmm 2/20/2005 17:18'!
isRunning

	^ socket isRunning! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 2/2/2005 12:26'!
materializeRequest: requestByteArray

	serializer resetOidManager.
	^ [ serializer materializeGraph: requestByteArray ]
		on: Error
		do:
			[ :exc |
			MaClientServerPreferences debug
				ifTrue: [ exc pass ]
				ifFalse:
					[ MaMalformedRequestError new
						messageText: exc messageText ;
						yourself ] ]! !

!MaTcpRequestServer methodsFor: 'start / stop' stamp: 'cmm 4/29/2003 23:28'!
processOn: portNumber
	"Listen on portNumber in the background.  For each client request received, evaluate my processBlock.  I queue all requests so that your block will never be evaluating more than one client-request at a time."

	socket server: self.
	socket listenOn: portNumber! !

!MaTcpRequestServer methodsFor: 'start / stop' stamp: 'cmm 2/20/2005 19:38'!
processOn: portNumber do: oneArgBlock
	"Listen on portNumber in the background.  For each client request received, evaluate oneArgBlock.  The argument to the block is the request object sent from the client.  I queue all requests so that your block will never be evaluating more than one client-request at a time."

self deprecated: 'Use processOn:using:'.
	processor _ oneArgBlock.
	self processOn: portNumber! !

!MaTcpRequestServer methodsFor: 'start / stop' stamp: 'cmm 2/20/2005 18:47'!
processOn: portNumber using: yourProcessor
	"Listen on portNumber in the background.  For each client request received, value: yourProcessor (which must respond to #value:).  I queue all requests and ensure your yourProcessor will never be evaluating more than one client-request at a time."

	processor _ yourProcessor.
	self processOn: portNumber! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 2/1/2005 15:51'!
processRequestByteArray: requestByteArray
	"Materialize the object serialized in requestByteArray, and serialize the response I get from telling my processor to process it."

	| requestObject responseObject |
	requestObject _ self materializeRequest: requestByteArray.
	responseObject _ requestObject = self connectionRequest
		ifTrue: [ self connectionInfo ]
		ifFalse: [ self processRequest: requestObject ].
	^ self serializeResponse: responseObject! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 6/22/2004 12:21'!
protocol: aCollection
	"Establish the class ids of classes in the protocol."

	aCollection do: [ : each | serializer classIdManager addNewClass: each ]! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 10/8/2002 22:20'!
protocolByteArray

	^serializer classDefinitionsByteArray! !

!MaTcpRequestServer methodsFor: 'private' stamp: 'cmm 12/27/2002 10:21'!
serializeResponse: anObject

	serializer resetOidManager.
	^(serializer serializeGraph: anObject) trimmedByteArray! !

!MaTcpRequestServer methodsFor: 'start / stop' stamp: 'cmm 12/31/2002 14:01'!
shutdown

	socket shutdown.
	"socket shutdown immediately disallows any more requests, but blocks until pending requests get processed.  Those remaining requests which are not finished writing responses can and will continue to write, but the server is no longer needed and able to shutdown."
	console noteServerShutdown! !

!MaTcpRequestServer methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:23'!
timeoutSeconds

	^socket timeoutSeconds! !

!MaTcpRequestServer methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:23'!
timeoutSeconds: anInteger

	socket timeoutSeconds: anInteger! !

!MaServer methodsFor: 'console' stamp: 'cmm 10/7/2002 23:45'!
console: aMaServerConsole

	console _ aMaServerConsole! !

!MaServer methodsFor: 'private' stamp: 'cmm 10/8/2002 23:05'!
initialize

	super initialize.
	console _ MaServerConsole new! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:29'!
note: aString
	"The socket is notifying us of a message."

	console note: aString! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:29'!
noteClientAborted: aSocket
	"Some client submitted aSocket to us, but when we tried to write to it, it was no longer connected."

	console noteClientAborted: aSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:29'!
noteException: anException whileReceivingRequestFrom: aSocket
	"anException was signaled when trying to read from aSocket.  Maybe they disconnected?"

	console 
		noteException: anException 
		whileReceivingRequestFrom: aSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:30'!
noteNoLongerAcceptingRequests
	"Called when the socket has previously been told to shutdown and can no longer accept requests."

	console noteNoLongerAcceptingRequests! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:30'!
noteNoLongerExtractingRequests
	"Called when the socket has previously been told to shutdown and extracted the last pending sockets request."

	console noteNoLongerExtractingRequests! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:30'!
noteNoLongerProcessingRequests
	"Called when the socket has previously been told to shutdown and has processed the last pending request."

	console noteNoLongerProcessingRequests! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 1/5/2003 23:03'!
noteRequestDequeued: requestByteArray for: aSocket
	"Indicates aSocket has delivered aByteArray as a request that it wants processed and the response returned."

	console
		noteRequestDequeued: requestByteArray
		for: aSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/16/2002 21:41'!
noteRequestQueued: aByteArray for: aSocket
	"Indicates aSocket has delivered aByteArray as a request that it wants processed and the response returned."

	console
		noteRequestQueued: aByteArray
		for: aSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:42'!
noteResponseSent: aByteArray to: aSocket
	"Successfully sent the aByteArray to aSocket."

	console
		noteResponseSent: aByteArray
		to: aSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:32'!
noteSendResponse: aByteArray timedOutOn: aSocket
	"aSocket must have disconnected or something, we couldn't get the response to them."

	console
		noteSendResponse: aByteArray
		timedOutOn: aSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:31'!
noteSendingResponse: aByteArray to: aSocket
	"a ByteArray is about to be sent to aSocket."

	console
		noteSendingResponse: aByteArray 
		to: aSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 1/5/2003 18:18'!
noteSocketDequeued: clientSocket
	"clientSocket has been accepted by the server, and a process forked to extracted its byteArray request."

	console noteSocketDequeued: clientSocket! !

!MaServer methodsFor: 'notifications' stamp: 'cmm 10/10/2002 00:32'!
noteSocketQueued: clientSocket
	"clientSocket has been accepted by the server, and a process forked to extracted its byteArray request."

	console noteSocketQueued: clientSocket! !

MaObject subclass: #MaServerConsole
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaServerConsole commentStamp: 'cmm 2/1/2005 15:34' prior: 0!
I am used to receive messages from the server.!

MaServerConsole subclass: #MaRequestServerConsole
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaRequestServerConsole commentStamp: 'cmm 2/1/2005 15:34' prior: 0!
I am used to receive messages from the server.!

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 1/5/2003 21:39'!
noteProcessingRequest: yourRequestObject
	"optional subclass responsibility"! !

!MaRequestServerConsole methodsFor: 'notifications' stamp: 'cmm 1/5/2003 22:41'!
noteRequestProcessed: yourRequestObject
	"optional subclass responsibility"! !

MaRequestServerConsole subclass: #MaTranscriptConsole
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaTranscriptConsole commentStamp: 'cmm 2/1/2005 15:41' prior: 0!
A default console that prints messages to the Transcript.!

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 10/1/2002 22:49'!
newLine

	Transcript cr! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/1/2002 23:33'!
note: aString

	self
		newLine;
		showMessage: aString! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 12/27/2004 00:05'!
noteException: anException onSocket: aSocket
	"anException was signaled when trying to read from aSocket.  Maybe they disconnected?"

	self
		newLine;
		showSocket: aSocket;
		showSeparator;
		showMessage:
			(anException messageText isEmpty
				ifTrue: [ anException class name ]
				ifFalse: [ anException messageText ])! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 12/27/2004 00:05'!
noteException: anException whileReceivingRequestFrom: aSocket
	"anException was signaled when trying to read from aSocket.  Maybe they disconnected?"

	self
		noteException: anException
		onSocket: aSocket! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/9/2002 23:07'!
noteNoLongerProcessingRequests
	"Called when the socket has previously been told to shutdown and has processed the last pending request."

	self newLine; showMessage: 'I shall process no more requests.'! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/1/2002 23:14'!
noteResponseSent: aByteArray to: aSocket
	"Successfully sent the aByteArray to aSocket."

	self
		newLine;
		showMessage: 'sent ', aByteArray size printString, ' bytes to ', aSocket remoteAddress maAsIpString! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/1/2002 23:15'!
noteSendResponse: aByteArray timedOutOn: aSocket
	"aSocket must have disconnected or something, we couldn't get the response to them."

	self
		newLine;
		showMessage: 'timed out trying to send ', aByteArray size printString, ' bytes to ', aSocket remoteAddress maAsIpString! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/1/2002 23:15'!
noteSendingResponse: aByteArray to: aSocket
	"a ByteArray is about to be sent to aSocket."

	self
		newLine;
		showMessage: 'about to send ', aByteArray size printString, ' bytes to ', aSocket remoteAddress maAsIpString! !

!MaTranscriptConsole methodsFor: 'notifications' stamp: 'cmm 10/1/2002 23:16'!
noteSocketQueued: clientSocket
	"clientSocket has been accepted by the server, and a process forked to extracted its byteArray request."

	self
		newLine;
		showMessage: 'queued ', clientSocket remoteAddress maAsIpString ! !

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 10/1/2002 22:49'!
showMessage: aString

	Transcript show: aString! !

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 10/1/2002 22:53'!
showSeparator

	Transcript show: ' - '! !

!MaTranscriptConsole methodsFor: 'output' stamp: 'cmm 10/1/2002 22:51'!
showSocket: aSocket

	Transcript show: aSocket remoteAddress maAsIpString! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
note: aString
	"The socket is notifying us of a message."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteException: anException whileReceivingRequestFrom: aSocket
	"anException was signaled when trying to read from aSocket.  Maybe they disconnected?"! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteNoLongerAcceptingRequests
	"Called when the socket has previously been told to shutdown and can no longer accept requests."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteNoLongerExtractingRequests
	"Called when the socket has previously been told to shutdown and extracted the last pending sockets request."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteNoLongerProcessingRequests
	"Called when the socket has previously been told to shutdown and has processed the last pending request."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 1/5/2003 22:53'!
noteRequestDequeued: requestByteArray for: aSocket
	"Indicates aSocket has delivered a ByteArray request that is about to be processed by my server."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/16/2002 21:41'!
noteRequestQueued: aByteArray for: aSocket
	"Indicates aSocket has delivered aByteArray as a request that it wants processed and the response returned."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteResponseSent: aByteArray to: aSocket
	"Successfully sent the aByteArray to aSocket."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteSendResponse: aByteArray timedOutOn: aSocket
	"aSocket must have disconnected or something, we couldn't get the response to them."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteSendingResponse: aByteArray to: aSocket
	"a ByteArray is about to be sent to aSocket."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/8/2002 23:10'!
noteServerShutdown! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 1/5/2003 18:18'!
noteSocketDequeued: clientSocket
	"clientSocket is about processed by the server."! !

!MaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 23:51'!
noteSocketQueued: clientSocket
	"clientSocket has been accepted by the server, and a process forked to extracted its byteArray request."
! !

MaObject subclass: #MaServerLink
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaServerLink commentStamp: 'cmm 2/1/2005 15:36' prior: 0!
Abstract superclass representing a clients link to a server.  A client uses this link to submit request objects via #submit:.!

MaServerLink subclass: #MaLocalServerLink
	instanceVariableNames: 'localRequestServer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaLocalServerLink commentStamp: 'cmm 2/1/2005 15:32' prior: 0!
This is your clients interface to the server when the server is running locally (e.g., with MaLocalRequestServer).!

!MaLocalServerLink class methodsFor: 'as yet unclassified' stamp: 'cmm 3/30/2002 22:50'!
localRequestServer: aMaLocalRequestServer

	^self new localRequestServer: aMaLocalRequestServer
! !

!MaLocalServerLink methodsFor: 'accessing' stamp: 'cmm 8/26/2003 22:50'!
hostAddress

	MaUserError signal: 'not remotly connected.'! !

!MaLocalServerLink methodsFor: 'testing' stamp: 'cmm 10/8/2002 23:59'!
isLocal

	^true! !

!MaLocalServerLink methodsFor: 'accessing' stamp: 'cmm 1/14/2005 15:11'!
linkInformation

	^ 'local-link to ', localRequestServer processor pathName! !

!MaLocalServerLink methodsFor: 'private' stamp: 'cmm 7/2/2004 13:17'!
localRequestServer

	^ localRequestServer! !

!MaLocalServerLink methodsFor: 'private' stamp: 'cmm 3/30/2002 22:52'!
localRequestServer: aMaLocalRequestServer

	localRequestServer _ aMaLocalRequestServer
! !

!MaLocalServerLink methodsFor: 'accessing' stamp: 'cmm 8/26/2003 22:50'!
port

	MaUserError signal: 'not remotly connected.'! !

!MaLocalServerLink methodsFor: 'actions' stamp: 'cmm 3/30/2002 22:03'!
submit: aMaServerRequest

	^localRequestServer processRequest: aMaServerRequest
! !

!MaServerLink methodsFor: 'actions' stamp: 'cmm 4/8/2002 23:33'!
connect
	"Remote links have connection code they need to run.."
! !

!MaServerLink methodsFor: 'actions' stamp: 'cmm 4/1/2002 00:25'!
disconnect
	"Remote links need to clean up resources."
! !

!MaServerLink methodsFor: 'testing' stamp: 'cmm 10/8/2002 23:59'!
isLocal

	^false! !

!MaServerLink methodsFor: 'override' stamp: 'cmm 7/2/2004 17:16'!
linkInformation

	self subclassResponsibility! !

!MaServerLink methodsFor: 'actions' stamp: 'cmm 3/30/2002 22:01'!
submit: aMaServerRequest

	self subclassResponsibility! !

!MaServerLink methodsFor: 'as yet unclassified' stamp: 'cmm 10/16/2002 22:05'!
timeoutSeconds: anInteger
	"for interface compatibility with remote server links."! !

MaServerLink subclass: #MaTcpRequestServerLink
	instanceVariableNames: 'serializer socket submitGuard'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaTcpRequestServerLink commentStamp: 'cmm 2/1/2005 15:40' prior: 0!
This is your clients interface to the server when the server is running remotely (e.g., with MaTcpRequestServer).!

!MaTcpRequestServerLink class methodsFor: 'creation' stamp: 'cmm 10/8/2002 22:17'!
hostAddress: aByteArray port: anInteger
	"bufferSize is the size of the byteArray used to send requests and materialize responses."

	^self new
		socket:
			(MaClientSocket
				hostAddress: aByteArray
				port: anInteger);
		yourself! !

!MaTcpRequestServerLink methodsFor: 'connections' stamp: 'cmm 1/2/2003 23:11'!
connect

	| idAndProtocol |
	idAndProtocol _ self submit: self connectionRequest.
	serializer classDefinitionsByteArray: (idAndProtocol at: 2)! !

!MaTcpRequestServerLink methodsFor: 'private' stamp: 'cmm 10/16/2002 14:26'!
connectionRequest

	^'connect'! !

!MaTcpRequestServerLink methodsFor: 'connections' stamp: 'cmm 4/15/2002 00:18'!
disconnect

	socket disconnect
! !

!MaTcpRequestServerLink methodsFor: 'private' stamp: 'cmm 10/14/2004 22:24'!
getByteArrayResponseFor: aMaServerRequest

	| graphBuffer |
	^ socket
		sendData: (graphBuffer _ serializer serializeGraph: aMaServerRequest) byteArray
		startingAt: 1
		count: graphBuffer bufferSize
		waitForReplyIn: serializer byteArray! !

!MaTcpRequestServerLink methodsFor: 'accessing' stamp: 'cmm 5/8/2002 23:24'!
hostAddress

	^socket hostAddress! !

!MaTcpRequestServerLink methodsFor: 'private' stamp: 'cmm 11/26/2004 13:15'!
initialize

	super initialize.
	self initializeSerializer.
	submitGuard _ Semaphore forMutualExclusion! !

!MaTcpRequestServerLink methodsFor: 'private' stamp: 'cmm 2/6/2005 12:37'!
initializeSerializer
	serializer _ MaObjectSerializer new.
	serializer preferences
		beforeSerializingAny: #Exception
		do:
			[ : anException |
			#('signalContext' 'handlerContext' 'outerContext') do:
				[ : each |
				anException
					instVarNamed: each
					put: nil ] ]! !

!MaTcpRequestServerLink methodsFor: 'accessing' stamp: 'cmm 7/2/2004 17:16'!
linkInformation

	^ (self hostAddress->self port) printString! !

!MaTcpRequestServerLink methodsFor: 'accessing' stamp: 'cmm 5/8/2002 23:24'!
port

	^socket port! !

!MaTcpRequestServerLink methodsFor: 'private' stamp: 'cmm 9/22/2002 21:32'!
socket: aMaClientSocket

	socket _ aMaClientSocket! !

!MaTcpRequestServerLink methodsFor: 'actions' stamp: 'cmm 3/12/2003 22:52'!
submit: aMaServerRequest

	| responseByteArray |
	^submitGuard critical:
		[ serializer resetOidManager.
		responseByteArray _ self getByteArrayResponseFor: aMaServerRequest.
		serializer resetOidManager.
		serializer materializeGraph: responseByteArray ]! !

!MaTcpRequestServerLink methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:23'!
timeoutSeconds

	^socket timeoutSeconds! !

!MaTcpRequestServerLink methodsFor: 'accessing' stamp: 'cmm 10/8/2002 22:22'!
timeoutSeconds: anInteger

	socket timeoutSeconds: anInteger! !

MaObject subclass: #MaServerSocket
	instanceVariableNames: 'socketQueue requestQueue socketExtractProcess processRequestProcess socket timeoutSeconds doBlock port running optimumRequestSize server'
	classVariableNames: 'ActiveSockets'
	poolDictionaries: ''
	category: 'Ma client server-API'!

!MaServerSocket commentStamp: 'cmm 2/1/2005 15:39' prior: 0!
I am the class which manages the underlying Squeak socket.  I receive and process requests in the order received in the background and send responses off in their own background process as well.

This class is based off of a design by John McIntosh.!

!MaServerSocket class methodsFor: 'initializing' stamp: 'cmm 4/30/2003 00:49'!
initialize

	ActiveSockets _ Set new.
	Smalltalk
		addToStartUpList: self;
		addToShutDownList: self! !

!MaServerSocket class methodsFor: 'initializing' stamp: 'cmm 12/27/2004 23:16'!
shutDown: aboutToQuit

	super shutDown: aboutToQuit.
	ActiveSockets do: [ :each | each primShutdown ]! !

!MaServerSocket class methodsFor: 'initializing' stamp: 'cmm 2/10/2005 09:42'!
startUp: amResuming

	super startUp: amResuming.
	ActiveSockets do: 
		[ :each | 
		"If another startUp method creates and starts one of me listening, then I don't want to tell it to listen again"
		each isRunning ifFalse: [ each listen ] ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 1/5/2003 22:52'!
dequeueByteArrayRequest

	| socketAndByteArrayRequest |
	socketAndByteArrayRequest _ requestQueue next.
	server 
		noteRequestDequeued: socketAndByteArrayRequest value
		for: socketAndByteArrayRequest key.
	^socketAndByteArrayRequest! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 10/3/2002 21:09'!
ensureByteArray: aByteArray canHold: anInteger

	| newAnswer |
	^aByteArray size < anInteger
		ifTrue:
			[ newAnswer _ ByteArray new: anInteger.
			newAnswer
				replaceFrom: 1
				to: aByteArray size
				with: aByteArray
				startingAt: 1.
			newAnswer ]
		ifFalse: [ aByteArray ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 1/5/2003 22:22'!
extractByteArrayFrom: aSocket

	[

	| requestByteArray |
	[ requestByteArray _ self receiveByteArrayFrom: aSocket ]
		on: Exception
		do:
			[ :exc | 
			server
				noteException: exc 
				whileReceivingRequestFrom: aSocket.
			nil ].
	requestByteArray ifNotNil: [ self queueRequest: aSocket->requestByteArray ]

		] fork! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 9/24/2002 23:01'!
headerSize
	"4-byte size header."

	^4! !

!MaServerSocket methodsFor: 'printing' stamp: 'cmm 10/2/2002 23:23'!
hostAndPortStringFor: aSocket

	^'host: ', aSocket remoteAddress maAsIpString, ' port: ', aSocket port printString! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 2/10/2005 09:45'!
initialize
	super initialize.
	running _ false.
	self initializeQueues.
	optimumRequestSize := 400.
	timeoutSeconds := 30.
	self server: MaServer new	"use the do-nothing server by default".
	doBlock _ [ :byteArray | server processRequestByteArray: byteArray ]! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 4/30/2003 00:30'!
initializeQueues

	socketQueue := SharedQueue new.
	requestQueue := SharedQueue new! !

!MaServerSocket methodsFor: 'testing' stamp: 'cmm 2/10/2005 09:42'!
isRunning

	^ running! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 2/10/2005 09:38'!
listen
	"By using this method, I will call my servers #processRequestByteArray:.  If you don't want to have a MaServer class, use listenOn:answer:."

	running ifTrue: [ MaClientServerUserError signal: 'socket already listening.' ].
	running _ true.
	ActiveSockets add: self.
	server note: 'initializing network drivers'.
	server note: 'creating server socket'.
	socket _ Socket newTCP.
	socket
		listenOn: port
		backlogSize: 50.
	socket isValid ifFalse: [ MaNetworkError signal: 'TCP/IP accept() not supported.' ].
	server note: 'now listening on ', port printString.
	self
		startAcceptingRequests;
		startExtractingRequests;
		startProcessingRequests! !

!MaServerSocket methodsFor: 'start/stop' stamp: 'cmm 4/30/2003 00:24'!
listenOn: portNumber
	"By using this method, I will call my servers #processRequestByteArray:.  If you don't want to have a MaServer class, use listenOn:answer:."

	port _ portNumber.
	self listen! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 6/30/2004 23:58'!
optimumRequestSize: anInteger

	optimumRequestSize _ anInteger! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 2/21/2005 17:28'!
primShutdown

	running _ false.
	server note: 'new sockets no longer accepted.'.
	socketQueue isEmpty ifFalse: [ server note: 'waiting for queued sockets..' ].
	[ socketQueue isEmpty ] whileFalse: [ (Delay forSeconds: 1) wait ].
	socketExtractProcess terminate.
	requestQueue isEmpty ifFalse: [ server note: 'waiting for pending requests to process..' ].
	[ requestQueue isEmpty ] whileFalse: [ (Delay forSeconds: 1) wait ].
	processRequestProcess terminate.
	server note: 'deallocating resources.'.
	socket ifNotNil: [ socket closeAndDestroy ].
	self initializeQueues.
	socket _ nil! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 9/29/2002 21:27'!
processRequest: socketAndRequestAssociation

	| responseByteArray |
	responseByteArray _ doBlock value: socketAndRequestAssociation value.
	self
		sendData: responseByteArray
		using: socketAndRequestAssociation key! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 1/5/2003 22:22'!
queueRequest: socketAndByteArray

	requestQueue nextPut: socketAndByteArray.
	server 
		noteRequestQueued: socketAndByteArray value 
		for: socketAndByteArray key! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 1/5/2003 18:13'!
queueSocket: aSocket

	socketQueue nextPut: aSocket.
	server noteSocketQueued: aSocket! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 7/2/2004 16:29'!
receiveByteArrayFrom: aSocket

	| bytesExpected bytesReceived answer |
	bytesReceived _ 0.
	bytesExpected _ nil.
	answer _ ByteArray new: optimumRequestSize.
	[ bytesExpected isNil or: [ bytesReceived < bytesExpected ] ]
		whileTrue:
			[ aSocket waitForDataFor: timeoutSeconds.
			aSocket dataAvailable ifFalse: [ MaNetworkError signal: 'data not available from ', (self hostAndPortStringFor: aSocket) ].
			bytesReceived _ bytesReceived
				+ (aSocket
					receiveSomeDataInto: answer
					startingAt: bytesReceived + 1).
			bytesExpected ifNil:
				[ bytesReceived >= self headerSize
					ifTrue:
						[ bytesExpected _ answer maUint: 32 at: 0.
						answer _ self
							ensureByteArray: answer 
							canHold: bytesExpected ] ] ].
	answer _ answer
		copyFrom: self headerSize + 1
		to: bytesExpected.
	^answer! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 12/27/2004 13:43'!
sendData: aByteArray using: aSocket

	[

	| byteArrayToSend |
	byteArrayToSend _ ByteArray new: aByteArray size + self headerSize.
	byteArrayToSend maUint: 32 at: 0 put: byteArrayToSend size.
	byteArrayToSend replaceFrom: self headerSize + 1 to: byteArrayToSend size with: aByteArray startingAt: 1.
	aSocket isConnected
		ifTrue:
			[ server noteSendingResponse: aByteArray to: aSocket.
			[ aSocket
				sendData: byteArrayToSend ;
				waitForSendDoneFor: timeoutSeconds ]
					on: ConnectionTimedOut
					do: [ : exc | aSocket closeAndDestroy ].
			(aSocket isConnected and: [ aSocket sendDone ])
				ifTrue:
					[ server noteResponseSent: aByteArray to: aSocket ]
				ifFalse:
					[ server noteSendResponse: aByteArray timedOutOn: aSocket ] ]
		ifFalse: [ server noteClientAborted: aSocket ].
	aSocket closeAndDestroy

		] forkAt: Processor userBackgroundPriority! !

!MaServerSocket methodsFor: 'server' stamp: 'cmm 12/30/2002 22:08'!
server: aMaServer
	"It is not required for the user to set this, a default one will be created."

	server _ aMaServer! !

!MaServerSocket methodsFor: 'start/stop' stamp: 'cmm 4/30/2003 00:46'!
shutdown

	ActiveSockets remove: self ifAbsent: [ "do nothing" ].
	self primShutdown! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 8/27/2004 12:53'!
startAcceptingRequests

	| clientSocket |
	[
		[ running ] whileTrue:
			[ clientSocket _ socket
				waitForAcceptFor: 5 "seconds"
				ifTimedOut: [ nil ].
			(clientSocket notNil and: [ clientSocket isConnected ])
				ifTrue: [ self queueSocket: clientSocket ] ].
		server noteNoLongerAcceptingRequests.

		] newProcess name: 'Socket acceptor'; resume.
	server note: 'Ready to queue sockets'! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 3/24/2005 12:22'!
startExtractingRequests

	socketExtractProcess _
		[

			[ running or: [ socketQueue notEmpty ] ]
				whileTrue:
					[ | clientSocket |
					clientSocket _ socketQueue next.
					server noteSocketDequeued: clientSocket.
					self extractByteArrayFrom: clientSocket.
					clientSocket _ nil ].
			server noteNoLongerExtractingRequests.

			] newProcess name: 'Request extractor'; resume.
	server note: 'ready to extract requests'! !

!MaServerSocket methodsFor: 'private' stamp: 'cmm 8/27/2004 12:54'!
startProcessingRequests

	processRequestProcess _ 

		[

			[ running or: [ requestQueue notEmpty ] ]
				whileTrue: [ self processRequest: self dequeueByteArrayRequest ].
			server noteNoLongerProcessingRequests.

			] newProcess name: 'Request processor'; resume.
	server note: 'ready to process requests'! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 9/29/2002 21:33'!
timeoutSeconds

	^timeoutSeconds! !

!MaServerSocket methodsFor: 'accessing' stamp: 'cmm 9/29/2002 21:32'!
timeoutSeconds: anInteger

	timeoutSeconds _ anInteger! !

PackageInfo subclass: #MaClientServerPackageInfo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Info'!

!MaClientServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 1/31/2003 00:20'!
additionalFiles

	^#('clientTest.ws' 'serverTest.ws')! !

!MaClientServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 4/8/2004 23:13'!
changesText

	^ '
version 5:
	updated for 3.6 and 3.7.  This version does not work with 3.5.

version 4:
	running servers now survive image restarts.  The network is initialized by MaClientSocket upon starting the image, allowing sockets to immediately reconnect upon image start up.

version 3:
  performance improvement
  support for multiple clients in the same image connected to the same remote server

version 2:
  Should now support big and little endian machines simultaneously.

version 1:
  Initial public release.'! !

!MaClientServerPackageInfo methodsFor: 'accessing' stamp: 'cmm 4/8/2004 23:11'!
prerequisitePackages
	^ #( 'Ma object serialization' )! !

!MaClientServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 1/30/2003 22:54'!
readmeText

	^'!!!!Ma client server

!!!!Introduction
I can help you write a client-server program very easily.

In the client-server model, a client sends a request to a server, the server processes the request and sends back a response.

This framework provides two layers interoperation.  You choose which layer you want to use and then write two programs.  A client program that will send requests to a remote server, which is running the other program to process those requests.

If you choose the lower layer, the "socket layer," you send ByteArray requests and receive ByteArray responses.  In the upper layer, the "object request layer" you send arbitrary request objects and get arbitrary objects as responses.  The "Ma object serialization" framework is used to convert the objects to ByteArrays which are sent and received for you using the socket layer.

!!!!Using the Socket Layer

!!Setting up the server
If you just need to send and receive ByteArray''s, you write your server program any way you want.  Just keep in mind that, somewhere, you''re going to want to take-in a ByteArray as input and you must answer a ByteArray as the response.

Then, to start the server:

=	myServerSocket _ MaServerSocket new
=		listenOn: 12345
=		answer:
=			[ :requestByteArray |
=			... give requestByteArray to your server program ... ]

MaServerSocket utilizes SharedQueue''s and background Processes to ensure that request processing is not held up by sending or receiving the ByteArray''s, which occurs entirely in the background.  The answer block is only ever evaluated for one client request at a time.

When you''re done, be sure to shutdown the socket so resources are properly deallocated:

=	myServerSocket shutdown

!!Getting answers in the client
Your client program creates a MaClientSocket by specifying the ip address as numbers in a ByteArray:

=	mySocket _ MaClientSocket hostAddress: #(192 168 1 1) port: 12345.

No connections are made just from creating the socket.  That happens when you actually send the request:

=	responseByteArray _ mySocket sendData: myRequestByteArray

There''s also a sendData:startingAt:count:waitForReplyIn: method for maximum efficiency.  If the reply ByteArray you supply wasn''t large enough, you''ll get a bigger one.

!!Socket Events
MaServerSocket knows about a "server", which defaults to a Server that does nothing (MaServer).  But you subclass MaServer and override some of its methods to learn when a socket is queued, a ByteArray request has been fully delivered, or sending a response, etc.  See MaServer for more information.

!!!!Using the Object Request Layer
If you choose to use the object request layer, you send "request" objects and get back arbitrary response objects.  So you write your client and server programs the same except you don''t concern yourself with parsing ByteArrays anymore.

!!Setting up the server
When you are ready to start your server program, instead of creating a MaServerSocket, you create a MaTcpRequestServer:

=	myServer _ MaTcpRequestServer protocol: { MyRequestClass1. MyRequestClass2.  MyResponseClass1.  MyResponseClass2.  OrderedCollection.  Dictionary. "etc." }.

What''s this?  Yes, you have to tell it all the classes that you''ll be sending and receiving, accounting for the full object-graph of your requests.  For example, if you have SubmitArticleRequest that has an Article object, you must include not only SubmitArticleRequest, Article and all the other kinds of objects that Article references, and what they reference, and so on.  You don''t however, need to ever specify String, ByteArray, Set, Symbol or Array.  Those are automatically part of every protocol.

You may also optionally specify a MaServerConsole to receive events.  Subclass MaServerConsole and override one of its notification methods and it''ll be called if you:

=	myServer _ console: myConsole

When you''re ready to start the server, you use processOn:do:

=	myServer
=		processOn: 12345
=		do:
=			[ :requestObject |
=			... pass requestObject to your server program, which answers the response ... ]

When you''re done, you need to shutdown to release operating system resources:

=	myServer shutdown

!!Setting up the client
Your client program creates a MaTcpRequestServerLink pointing to the server hostAddress and port is is listening:

=	myLinkToServer _ MaTcpRequestServerLink
=		hostAddress: #(192 168 1 1)
=		port: 12345

Before you can do anything, you have to connect.  This will ensure you have all the needed protocol in this client image (and throw an exception if you don''t).

=	myLinkToServer connect

To balance network fluctuations against your patience, you may wish to override the default timeout period from 30 seconds:

=	myLinkToServer timeoutSeconds: 30

!!Getting answers in the client
Now that you''re connected, your client program can get answers from the server by:

=	myLinkToServer submit: someRequestObject

Remember, someRequestObject will be a member of the protocol, as will the response you get back from calling submit:.

That''s it!!  When you''re done, don''t forget to disconnect so as you release network resources back to the operating system.

=	myLinkToServer disconnect
'! !

MaError subclass: #MaMalformedRequestError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma client server-Exceptions'!

!MaMalformedRequestError commentStamp: 'cmm 2/1/2005 15:26' prior: 0!
I am signaled when a request is received from a client that could not be materialized.!
MaClientServerPreferences initialize!
MaClientSocket initialize!
MaServerSocket initialize!
